<?php 
/**
###############################
#                             #
#  ████  █████   ████  █████  #
# ██  ██ ██  ██   ██   ██  ██ #
# ██     ██  ██   ██   ██  ██ #
# ██ ███ █████    ██   ██  ██ #
# ██  ██ ██  ██   ██   ██  ██ #
# ██  ██ ██  ██   ██   ██  ██ #
#  ████  ██  ██  ████  █████  #
#                             #
###############################
by bronco@warriordudimanche.net
v1.2 bugfix on overwriting some letters
 */
class grid
{
	private $grid;
	private $words;
	private $cellsBySide=15; # max working area (cropped after)
	private $horizontalOnly=false;
	private $capitalForWordsOnly=false;
	private $reverse=false;

	private $currentLine=0;
	private $currentColumn=0;
	private $tempLinesArray=[];
	private $tempColumnsArray=[];

	function __construct($words=null,$width=null)
	{
		# init
		if ($width){
			$this->setCellsBySide($width);
		}
		if ($words){
			$this->setWords($words);
		}
	}


	function placeWords(){
		if (empty($this->words)){return false;}

		shuffle($this->words);
		foreach ($this->words as $nb=>$word) {
			$this->resetTempArrays();
			
			if ($this->reverse){
				if (rand(0,1)==1){
					$word=strrev($word);
				}
			}

			if ( $this->horizontalOnly || rand(0,1)==0 ){
				$done=false;
				while ($done==false && $this->currentLine!==false){
					$this->currentLine=$this->randomLine();
					$done=$this->placeInLine($word);
				}

				$this->currentColumn=0;

			}else{
				$done=false;
				while ($done==false && $this->currentColumn<$this->cellsBySide){
					$done=$this->placeInColumn($word);
					$this->currentColumn=$this->randomColumn();
				}
				$this->currentLine=0;
			}
			
			
		}

		$this->fillEmptyCells();
		return $this->grid;
	}

	function placeInLine(string $word){
		$length=strlen($word);
		$free_spaces=[0];
		$current_part_index=0;
		# check & measure all free spaces chunks in the current line 
		foreach ($this->grid[$this->currentLine] as $char_position => $cell_content) {
			if (!$cell_content){

				if (empty($free_spaces[$current_part_index])){
					$free_spaces[$current_part_index]=1;
				}else{
					$free_spaces[$current_part_index]++;	
				}
			}else{
				$current_part_index=$char_position+1;
			}
		}

		foreach ($free_spaces as $column => $nb) {
			if ($nb>=$length){
				# if current word fits in current free space part
				$margin=$nb-$length;
				if ($margin>0){
					$first_word_cell=rand($column,$column+$margin);
				}else{
					$first_word_cell=$column;
				}
				$last_word_cell=$first_word_cell+$length;

				# place word's letters in each free space's cell
				for ($cell=$first_word_cell;$cell<$last_word_cell;$cell++){
					$this->grid[$this->currentLine][$cell]=$word[$cell-$first_word_cell];
				}
				return true;
			}
		}
		return false;
	}



	function placeInColumn(string $word){
		$length=strlen($word);
		$free_spaces=[0];
		$current_part_index=0;
		# check all free spaces in the current column
		for ($line=0;$line<$this->cellsBySide;$line++){
			$cell_content=$this->grid[$line][$this->currentColumn];
			if (!$cell_content){
				if (empty($free_spaces[$current_part_index])){
					$free_spaces[$current_part_index]=1;
				}else{
					$free_spaces[$current_part_index]++;	
				}
			}else{
				$current_part_index=$line+1;
			}
		}
		foreach ($free_spaces as $key => $nb) {
			if ($nb>=$length){
				# if current word fits in current free space part
				$margin=$nb-$length;
				if ($margin>0){
					$first_word_cell=rand($key,$key+$margin);
				}else{
					$first_word_cell=$key;
				}
				$last_word_cell=$first_word_cell+$length;

				for ($line=$first_word_cell;$line<$last_word_cell;$line++){
					$this->grid[$line][$this->currentColumn]=$word[$line-$first_word_cell];
				}
				return true;
			}
		}
		return false;
		
	}

	function resetTempArrays(){
		$this->tempLinesArray=$this->tempColumnsArray=array_fill(0, $this->cellsBySide, true);
	}

	function randomLine(){
		if (count($this->tempLinesArray)>0){
			$random_index=array_rand($this->tempLinesArray);
			unset($this->tempLinesArray[$random_index]);
			return $random_index;
		}else{
			return false;
		}
	}

	function randomColumn(){
		if (count($this->tempColumnsArray)>0){
			$random_index=array_rand($this->tempColumnsArray);
			unset($this->tempColumnsArray[$random_index]);
			return $random_index;
		}else{
			return false;
		}
	}

	function doEmptyGrid(){
		$this->grid=array_fill(0,$this->cellsBySide , array_fill(0,$this->cellsBySide , false));
	}
	function fillEmptyCells(){
		foreach ($this->grid as $line => $line_content) {
			foreach ($line_content as $column => $cell_content) {
				if (!$cell_content){
					$this->grid[$line][$column]=$this->randomLetter();
				}
			}
		}
	}

	function randomLetter(){return !$this->capitalForWordsOnly?chr(rand(65,90)):chr(rand(97,122));}


	function getGrid(){return $this->grid;}
	function setHorizontalOnly($tf){$this->horizontalOnly=$tf;}
	function setCapitalForWordsOnly($tf){$this->capitalForWordsOnly=$tf;}
	function setReverse($tf){$this->reverse=$tf;}
	function setCellsBySide($nb){
		$this->cellsBySide=$nb;
	}
	
	function setWords($words){
		$this->words=explode(',',strtoupper($words));
		$max_word_length=0;
		foreach ($this->words as $word) {
			$max_word_length=max($max_word_length,strlen($word));
		}
		$this->cellsBySide=max( 
			count($this->words),
			$max_word_length+2,
			$this->cellsBySide,
		);

		$this->doEmptyGrid();
	}

	function echoGrid($before='<pre>',$after='</pre>',$line_separator="<br>\n",$letter_separator='&nbsp;&nbsp;&nbsp;'){
		echo $before;
		foreach ($this->grid as $key => $value) {
		 	echo implode($letter_separator, $value).$line_separator;
		} 
		echo $after;
	}

	function grid2lines(){
		$array=[];
		foreach ($this->grid as $key => $value) {
		 	$array[$key]= implode(' ', $value);
		} 
		return $array;
	}
}